/*
 * This program is based on two examples provided by Greg Kroah-Hartman:
 *
 *   USB LED driver - 1.1
 *   Copyright (C) 2004 Greg Kroah-Hartman (greg@kroah.com)
 *
 *   USB Skeleton driver - 2.0
 *   Copyright (C) 2001-2004 Greg Kroah-Hartman (greg@kroah.com)
 *
 *        This program is free software; you can redistribute it and/or
 *        modify it under the terms of the GNU General Public License as
 *        published by the Free Software Foundation, version 2.
 *
 * This document by Patrick Mochel was also very helpful:
 * http://www.mjmwired.net/kernel/Documentation/filesystems/sysfs.txt
 *
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/usb.h>

#define DRIVER_AUTHOR "J. Mike Rollins, rollins@CamoTruck.Net"
#define DRIVER_DESC "USB reping driver"

#define VENDOR_ID	0x04d8
#define PRODUCT_ID	0x000c

#define USBGEN_EP_SIZE  64
#define KHz  *1000

/* table of devices that work with this driver */
static struct usb_device_id id_table [] = {
	{ USB_DEVICE(VENDOR_ID, PRODUCT_ID) },
	{ },
};
MODULE_DEVICE_TABLE (usb, id_table);

/* Structure to hold all of our device specific stuff */
struct usb_reping {
        struct usb_device *     udev;                   /* the usb device for this device */
        struct usb_interface *  interface;              /* the interface for this device */
        size_t                  bulk_in_size;           /* the size of the receive buffer */
        __u8                    bulk_in_endpointAddr;   /* the address of the bulk in endpoint */
        __u8                    bulk_out_endpointAddr;  /* the address of the bulk out endpoint */
        char                    out_string[USBGEN_EP_SIZE];         
        char                    in_string[USBGEN_EP_SIZE];         
};

/* show_reping is called when the sysfs file is read.
 * We should display the contents from the out buffer.  
 * The out buffer is populated from the PIC18F4550 chip. 
 */
static ssize_t show_reping(struct device *dev, struct device_attribute *attr, char *buf)
  {
    struct usb_interface *intf = to_usb_interface(dev);
    struct usb_reping *reping = usb_get_intfdata(intf);

    // NOTE: buf is PAGE_SIZE (about 4K)
    return sprintf(buf, "%s\n", reping->out_string);
  }

/* store_reping is called when the sysfs file is written.
 * The contents from the buf parameter will be sent via usb to the PIC18F4550 chip. 
 * A bulk read then retrieves the data from the chip (hopefully in the reverse order).
 * The new data is stored in the out_string buffer.
 */
static ssize_t store_reping(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
  {
    struct usb_interface *intf = to_usb_interface(dev);
    struct usb_reping *reping = usb_get_intfdata(intf);

    int i,bytes,retval = 0;
    char *buffer;
    
    buffer = kzalloc(USBGEN_EP_SIZE, GFP_KERNEL);
    if (! buffer)
      {
        dev_info(dev, "Could not allocate bulk_in_buffer");
        goto error;
      }

    if (count > (USBGEN_EP_SIZE-1))
      {
        dev_info(dev, "string data too long");
        goto error;
      }

    /* clean out our local copy */
    memset(reping->in_string, 0, USBGEN_EP_SIZE);

    /* 
     * buffer[0] is the size of the text to send to the PIC18F4550 chip
     * buffer[1..count] is the text to send to the PIC18F4550 chip
     */
    buffer[0] = count;

    for (i=0; i<count; i++)
        buffer[i+1] = reping->in_string[i] = buf[i];

    /* do a blocking bulk write to send data to the device */
    retval = usb_bulk_msg(reping->udev,
                          usb_sndbulkpipe(reping->udev, reping->bulk_out_endpointAddr),
                          buffer,
                          (count+1),  // [len][text]
                          &bytes, 
                          10 KHz);

    if (retval)
      {
        dev_info(dev, "usb_bulk_msg write returned an error: %d",retval);
        retval = -EFAULT;
        goto error;
      }

    if ( !(bytes == (count+1)) )
      {
        dev_info(dev, "usb_bulk_msg write byte count mismatch (%d != %d)",bytes,(count+1));
        retval = -EFAULT;
        goto error;
      }

    /* clean out our buffer */
    memset(buffer, 0, USBGEN_EP_SIZE);
    bytes = 0;

    /* do a blocking bulk read to get data from the device */
    retval = usb_bulk_msg(reping->udev,
                          usb_rcvbulkpipe(reping->udev, reping->bulk_in_endpointAddr),
                          buffer,
                          (count+1),  // [len][text]
                          &bytes, 
                          10 KHz);

    if (retval)
      {
        dev_info(dev, "usb_bulk_msg read returned an error: %d",retval);
        retval = -EFAULT;
        goto error;
      }

    if ( !(bytes == (count+1)) )
      {
        dev_info(dev, "usb_bulk_msg read byte count mismatch (%d != %d)",bytes,(count+1));
        retval = -EFAULT;
        goto error;
      }

    if ( !(buffer[0] == count) )
      {
        dev_info(dev, "usb_bulk_msg read string length incorrect, %d.)",buffer[0]);
        retval = -EFAULT;
        goto error;
      }

    /* clean out our local copy */
    memset(reping->out_string, 0, USBGEN_EP_SIZE);

    for (i=0; i<buffer[0]; i++)
        reping->out_string[i] = buffer[i+1];

    retval = count;

error:
    kfree(buffer);
    return retval;
  }


static struct device_attribute dev_attr_reping = {
       .attr	= {
		.name = "reping",
		.mode = S_IWUSR | S_IRUGO,
	},
	.show = show_reping,
	.store = store_reping,
};


static int reping_probe(struct usb_interface *interface, const struct usb_device_id *id)
  {
    struct usb_device *udev = interface_to_usbdev(interface);
    struct usb_reping *dev = NULL;
    int retval = -ENOMEM;

    /* usb setup */
    struct usb_host_interface *iface_desc;
    struct usb_endpoint_descriptor *endpoint;
    size_t buffer_size;
    int i;

    dev = kzalloc(sizeof(struct usb_reping), GFP_KERNEL);
    if (dev == NULL) 
      {
        dev_err(&interface->dev, "Out of memory\n");
        goto error;
      }

    dev->udev = usb_get_dev(udev);

    /* sysfs setup */
    device_create_file(&interface->dev, &dev_attr_reping);

    /* usb setup */
    dev->interface = interface;

    /* set up the endpoint information */
    /* use only the first bulk-in and bulk-out endpoints */
    iface_desc = interface->cur_altsetting;
    for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) 
      {
        endpoint = &iface_desc->endpoint[i].desc;

        /* check for endpoints that are IN and BULK */
        if (!dev->bulk_in_endpointAddr &&
            ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
                                    == USB_DIR_IN) &&
            ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
                                    == USB_ENDPOINT_XFER_BULK)) 
          {
            /* we found a bulk in endpoint */
            buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);
            dev->bulk_in_size = buffer_size;
            dev->bulk_in_endpointAddr = endpoint->bEndpointAddress;
          }


        /* check for endpoints that are OUT and BULK */
        if (!dev->bulk_out_endpointAddr &&
            ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
                                    == USB_DIR_OUT) &&
            ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
                                    == USB_ENDPOINT_XFER_BULK)) 
          {
            /* we found a bulk out endpoint */
            dev->bulk_out_endpointAddr = endpoint->bEndpointAddress;
          }
      } // for 

      if (!(dev->bulk_in_endpointAddr && dev->bulk_out_endpointAddr)) 
        {
          dev_info(&interface->dev, "Could not find both bulk-in and bulk-out endpoints");
          goto error;
        }

    /* save our data pointer in this interface device */
    usb_set_intfdata (interface, dev);

    /* let the user know what node this device is now attached to */
    dev_info(&interface->dev, 
        "USB reping device now attached: EP_IN %d; EP_OUT %d; SIZE %d\n",
        dev->bulk_in_endpointAddr, 
        dev->bulk_out_endpointAddr, 
        dev->bulk_in_size);

    return 0;

error:
    kfree(dev);
    return retval;
}


static void reping_disconnect(struct usb_interface *interface)
  {
    struct usb_reping *dev;

    dev = usb_get_intfdata (interface);
    usb_set_intfdata (interface, NULL);
    
    device_remove_file(&interface->dev, &dev_attr_reping);
    
    usb_put_dev(dev->udev);
    
    kfree(dev);
    
    dev_info(&interface->dev, "USB reping now disconnected\n");
  }

static struct usb_driver reping_driver = {
	.name =		"reping",
	.probe =	reping_probe,
	.disconnect =	reping_disconnect,
	.id_table =	id_table,
};

static int __init usb_reping_init(void)
  {
    int retval = 0;

    retval = usb_register(&reping_driver);
    if (retval)
        err("usb_register failed. Error number %d", retval);
    return retval;
  }

static void __exit usb_reping_exit(void)
  {
    usb_deregister(&reping_driver);
  }

module_init (usb_reping_init);
module_exit (usb_reping_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
